/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/
//#define DEBUG

#include "application.h"
#include "ble_api.h"
#include "ble_api_master.h"
#include "ble_api_client.h"
#include "bt_api_ven_flash.h"
#include "client_generic_access_api.h"
#include "client_device_information_api.h"
#include "crc8.h"
#include "debug_api.h"

#include "host_api.h"

/*! Structure that contains central-specific application data */
struct data {
  bool is_scanning;                    /*!< Indicates if the device is currently scanning */
  bool is_name_found;                  /*!< Indicates if the correct name has been found */
  bool is_uuid_found;                  /*!< Indicates if the correct UUID has been found */
  uint8_t bd_addr[6];                  /*!< Bluetooth device address of device to connect to */
  uint16_t connection_handle;          /*!< connection handle of the currently active connection */
  enum tcu_le_smp_address_type addr_type; /*!< Address type of device to connect to */
};
static struct data data;

/*! Structure that contains configuration data for the central */
struct config {
  uint8_t device_name[MAX_DEVICE_NAME_LEN + 1]; /*!< device name used for advertising */
  uint8_t device_name_len;                      /*!< length of device name */
  uint8_t crc8;                                 /*!< crc value - must be last */
};
static struct config config;

/* Helper function that handles failure of most operations by disconnecting from the remote device */
static void fail_handler(void)
{
  /* todo: implement disconnect */
  /* todo: turn led off for sure */
  /* todo: start scanning again */
}

/* Callback function that is called whenever enabling or disabling of  the client characteristic configuration descriptor of the ring characteristic has finished */
static void characteristic_ring_characteristic_descriptor_enable_disable_notification_indication_status(uint8_t operation, uint8_t status, uint16_t connection_handle)
{
  /* something went wrong */
  if (status) {
    fail_handler();
    return;
  }

  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called whenever a write to the client characteristic configuration descriptor of the ring characteristic has finished */
static void characteristic_ring_characteristic_descriptor_write_done(uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called whenever a write to the client characteristic configuration descriptor of the ring characteristic has failed */
static void characteristic_ring_characteristic_descriptor_write_failed(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);

  /* something went wrong */
  fail_handler();
}

/* Declaration of storage for runtime data of the characteristic client characteristic configuration descriptor for the ring characteristic */
static struct ble_api_cli_characteristic_descriptor_ram characteristic_ring_characteristic_descriptor_ram[MAXIMUM_DEVICES];

/* Declaration of the the characteristic client characteristic configuration descriptor for the ring characteristic  */
static struct ble_api_cli_characteristic_descriptor characteristic_ring_characteristic_descriptor = {
    {
        LE_SHORT_UUID, SET_LE_SHORT_UUID(0x2902), characteristic_ring_characteristic_descriptor_enable_disable_notification_indication_status, characteristic_ring_characteristic_descriptor_write_done, characteristic_ring_characteristic_descriptor_write_failed, NULL, /* read_done */
        NULL,                                                                                                                                                                                                                                                              /* read_failed */
    },
    &characteristic_ring_characteristic_descriptor_ram[0],
};

/* Array of all characterstic descriptors for the ring characteristic - there is only one */
static struct ble_api_cli_characteristic_descriptor *characteristic_ring_characteristic_descriptors[] = {
    &characteristic_ring_characteristic_descriptor,
};

/* Total number of characterstic descriptors for the ring characteristic */
#define NUM_CHARACTERISTIC_RING_CHARACTERISTIC_DESCRIPTORS (sizeof(characteristic_ring_characteristic_descriptors) / sizeof(characteristic_ring_characteristic_descriptors[0]))

/* Callback function that is called whenever an indication for the ring characteristic has been received */
static enum ble_api_result characteristic_ring_indication(uint8_t *value, uint16_t count, uint16_t connection_handle)
{
  /* we only accept ring indications that contain a single byte */
  if (count != 1) {
    return LE_API_ERROR;
  }

  /* depending on the value that was transmitted issue a diagnostic message to the user */
  if (*value) {
    xprintf("RINGTONE START\n");
  } else {
    xprintf("RINGTONE STOP\n");
  }

  dprintf("value %02x\n", *value);
  return LE_API_SUCCESS;
}

/* Callback function that is called whenever a write to the ring characteristic has finished */
static void characteristic_ring_write_done(uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called whenever a write to the ring characteristic has failed */
static void characteristic_ring_write_failed(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);

  /* something went wrong */
  fail_handler();
}

/* Callback function that is called whenever a read from the ring characteristic has finished */
static void characteristic_ring_read_done(uint8_t *data, uint16_t length, uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called whenever a write to the ring characteristic has failed */
static void characteristic_ring_read_failed(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);

  /* something went wrong */
  fail_handler();
}

/* Declaration of storage for runtime data of the ring characteristic */
static struct ble_api_cli_characteristic_ram characteristic_ring_ram[MAXIMUM_DEVICES];

/* Declaration of the ring characteristic */
static struct ble_api_cli_characteristic characteristic_ring_definition = {
    {
        LE_LONG_UUID,
        /* UUID 49700b78-1e46-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
            0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x46, 0x1e, 0x78, 0x0b, 0x70, 0x49,
            // clang-format on
        },
        &characteristic_ring_characteristic_descriptors[0],
        NUM_CHARACTERISTIC_RING_CHARACTERISTIC_DESCRIPTORS,
        NULL, /* notification */
        characteristic_ring_indication,
        characteristic_ring_write_done,
        characteristic_ring_write_failed,
        characteristic_ring_read_done,
        characteristic_ring_read_failed,
    },
    &characteristic_ring_ram[0],
};

/* "Current tune" characteristic */

/* Callback function that is called when a write to the current tune characteristic value was successful */
static void characteristic_current_tune_write_done(uint16_t connection_handle)
{
  /* this is the confirmation to the user that the write to the remote device was successful */
  xprintf("OK\n");

  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called when a write to the current tune characteristic value failed */
static void characteristic_current_tune_write_failed(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  /* this is the confirmation to the user that the write to the remote device has failed - maybe there is no connection? */
  xprintf("FAILED\n");

  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);
}

/* Callback function that is called when a read from the current tune characteristic value was successful */
static void characteristic_current_tune_read_done(uint8_t *data, uint16_t length, uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
}

/* Callback function that is called when a read from the current tune characteristic value failed */
static void characteristic_current_tune_read_failed(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);
}

/* Declaration of storage for runtime data of the current tune characteristic */
static struct ble_api_cli_characteristic_ram characteristic_current_tune_ram[MAXIMUM_DEVICES];

/* Declaration of the current tune characteristic */
static struct ble_api_cli_characteristic characteristic_current_tune_definition = {
    {
        LE_LONG_UUID,
        /* UUID 82ab3028-1e48-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
          0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x48, 0x1e, 0x28, 0x30, 0xab, 0x82,
            // clang-format on
        },
        NULL, /* descriptors */
        0,    /* num_descriptors */
        NULL, /* notification */
        NULL, /* indication */
        characteristic_current_tune_write_done,
        characteristic_current_tune_write_failed,
        characteristic_current_tune_read_done,
        characteristic_current_tune_read_failed,
    },
    &characteristic_current_tune_ram[0],
};

/* Doorbell service */

/* Declaration of all the characteristics of the doorbell service */
static struct ble_api_cli_characteristic *service_doorbell_characteristics[] = {
    &characteristic_ring_definition,
    &characteristic_current_tune_definition,
};

/* Total number of services of the doorbell service */
#define NUM_DOORBELL_SERVICE_CHARACTERISTICS sizeof(service_doorbell_characteristics) / sizeof(service_doorbell_characteristics[0])

/* Declaration of storage for runtime data of the doorbell service  */
static struct ble_api_cli_service_ram service_doorbell_ram[MAXIMUM_DEVICES];

/* Declaration of the doorbell service */
static struct ble_api_cli_service client_service_doorbell = {
    {
        LE_LONG_UUID,
        /* UUID dcace68e-1e48-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
          0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x48, 0x1e, 0x8e, 0xe6, 0xac, 0xdc,
            // clang-format on
        },
        &service_doorbell_characteristics[0],
        NUM_DOORBELL_SERVICE_CHARACTERISTICS,
        NULL, /* found_callback, */
    },
    &service_doorbell_ram[0],
};

/* Callback function that is called when the discover process is finished */
static void discover_done_callback(uint16_t connection_handle, uint16_t mtu)
{
  enum ble_api_result result;
  int ret;

  /* this function must decide if the information found is enough to say that the profile exists */

  if ((!client_service_doorbell.ram[connection_handle - 1].service_handle) || (!client_service_doorbell.ram[connection_handle - 1].end_group_handle)) {
    fail_handler();
    return;
  }
  dprintf("doorbell service; start handle: 0x%04x, end handle: 0x%04x\n", client_service_doorbell.ram[connection_handle - 1].service_handle, client_service_doorbell.ram[connection_handle - 1].end_group_handle);

  if (!characteristic_ring_definition.ram[connection_handle - 1].descriptor_handle) {
    fail_handler();
    return;
  }
  dprintf("characteristic ring declaration handle: 0x%04x\n", characteristic_ring_definition.ram[connection_handle - 1].descriptor_handle);

  if (!characteristic_ring_definition.ram[connection_handle - 1].value_handle) {
    fail_handler();
    return;
  }
  dprintf("characteristic ring value handle: 0x%04x\n", characteristic_ring_definition.ram[connection_handle - 1].value_handle);

  if ((characteristic_ring_definition.ram[connection_handle - 1].properties) != (LE_WRITE | LE_READ | LE_INDICATE)) {
    fail_handler();
    return;
  }
  dprintf("supported characteristic ring characteristic property: 0x%02x\n", characteristic_ring_definition.ram[connection_handle - 1].properties);

  if (!characteristic_current_tune_definition.ram[connection_handle - 1].descriptor_handle) {
    fail_handler();
    return;
  }
  dprintf("characteristic current_tune declaration handle: 0x%04x\n", characteristic_current_tune_definition.ram[connection_handle - 1].descriptor_handle);

  if (!characteristic_current_tune_definition.ram[connection_handle - 1].value_handle) {
    fail_handler();
    return;
  }
  dprintf("characteristic current_tune value handle: 0x%04x\n", characteristic_current_tune_definition.ram[connection_handle - 1].value_handle);

  if ((characteristic_current_tune_definition.ram[connection_handle - 1].properties) != (LE_WRITE | LE_READ)) {
    fail_handler();
    return;
  }
  dprintf("supported characteristic ring characteristic property: 0x%02x\n", characteristic_current_tune_definition.ram[connection_handle - 1].properties);

  data.connection_handle = connection_handle;

  /* LED is lit when connection is established and profile has been found */
  ret = Driver_GPIO.Write(GPIO10, HIGH);
  if (ret) {
    warning("Driver_GPIO.Write() failed\n");
  }

  /* Enable indications on the ring characteristic */
  result = ble_api_cli_characteristic_enable_indication_start(&characteristic_ring_definition, connection_handle);
  if (result) {
    dprintf("ble_api_cli_characteristic_enable_indiaction_start() failed\n");
    fail_handler();
    return;
  }

  xprintf("CONNECT " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_BE(data.bd_addr));

  dprintf("connection_handle %04x, mtu %d\n", connection_handle, mtu);
}

/* Callback function that is called when the discover process has failed */
static void discover_failed_callback(uint16_t connection_handle, uint16_t mtu)
{
  /* profile not found handling */
  fail_handler();

  dprintf("\n");
}

/* Helper function that starts a scan for other devices */
static int scan_start(void)
{
  enum ble_api_result result;

  struct ble_api_uuid uuid = {client_service_doorbell.rom.uuid_type};
  memcpy(&uuid.uuid, client_service_doorbell.rom.uuid, LE_LONG_UUID);

  static struct ble_api_legacy_scan_parameters legacy_scan_parameters = {
      0x20, // LE_SCAN_INTERVAL,
      0x20, // LE_SCAN_WINDOW,
      LE_SCAN_FILTER_POLICY,
      0x00, // LE_SCAN_DUPLICATION_FILTER
      LE_ACTIVE,
  };

  /* filter for all devices and do the filtering for both
  the UUID and the name manually later. this is necessary, because UUID and
  device name may arrive in the advertising data or the scan response
  separately... */

  result = ble_api_configure_legacy_scan_parameters(&legacy_scan_parameters);
  if (result != LE_API_SUCCESS) {
    return -1;
  }

  result = ble_api_master_start_scan_filter_none(ADDRESS_CONFIGURATION_PUBLIC);
  if (result) {
    dprintf("ble_api_master_start_scan_filter_none() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

/* Callback function that is called when the connection is terminated */
static void disconnected_callback(const struct ble_api_cli_profile *profile, uint16_t connection_handle)
{
  int ret;

  dprintf("connection_handle %04x\n", connection_handle);

  ret = Driver_GPIO.Write(GPIO10, LOW);
  if (ret) {
    warning("Driver_GPIO.Write() failed\n");
  }

  data.is_uuid_found = false;
  data.is_name_found = false;

  /* immediately start scanning again - maybe to re-establish a lost connection */
  ret = scan_start();
  if (ret) {
    dprintf("scan_start() failed\n");
    return;
  }

  xprintf("DISCONNECT\n");
}

/* Doorbell profile */

/* Declaration of all the services of the doorbell profile */
static struct ble_api_cli_service *services[] = {
    &client_service_generic_access,
    &client_service_device_information,
    &client_service_doorbell,
};

/* Declaration of storage for runtime data of the doorbell profile  */
static struct ble_api_cli_profile_ram doorbell_profile_ram[MAXIMUM_DEVICES];

/* Declaration of the doorbell profile */
// clang-format off
static struct ble_api_cli_profile doorbell_profile = {
    {
      services, sizeof(services) / sizeof(services[0]), 
      SMP_AUTH_NO_MITM_NO_BONDING, 
      discover_done_callback, 
      discover_failed_callback, 
      disconnected_callback
    },
    &doorbell_profile_ram[0],
};
// clang-format on

/* Helper function that updates the configuration data in the flash */
static int write_config(void)
{
  int ret;

  /* Calculate a crc for data integrity */
  config.crc8 = crc((void *)&config, sizeof(config) - 1);

  ret = application_write_config(&config, sizeof(config));
  if (ret) {
    dprintf("application_write_config() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

/* Helper function that reads the configuration data from flash */
static int read_config(void)
{
  /* The default device name to use when no configuration is present */
  static const uint8_t device_name[] = "PAN1760A\0";
  uint8_t calculated_crc8;
  int ret;

  ret = application_read_config(&config, sizeof(config));
  if (ret) {
    dprintf("application_read_config() failed\n");
    return -1;
  }

  /* check if configuration data is valid by comparing a calculated crc with the stored one */
  calculated_crc8 = crc((void *)&config, sizeof(config) - 1);
  if (calculated_crc8 == config.crc8) {
    /* do nothing - config is valid */
    return 0;
  }

  /* use default configuration */
  memcpy(&config.device_name[0], &device_name[0], sizeof(device_name));
  config.device_name_len = sizeof(device_name) - 1; /* do not count trailing \0 */

  dprintf("\n");
  return 0;
}

/* Callback function that is called in application context to indicate that the button has been pressed */
void button_handler_callback(void)
{
  uint8_t buf = 0x01;
  enum ble_api_result result;

  /* write any value to the ring characteristic to trigger a ring on the remote device */
  result = ble_api_cli_characteristic_write_value_start(&characteristic_ring_definition, &buf, sizeof(buf), data.connection_handle);
  if (result) {
    dprintf("ble_api_cli_characteristic_write_value_start() failed\n");
    return;
  }

  dprintf("\n");
}

/* Low-level button interrupt handler - called from ISR context */
static void button_interrupt(uint32_t arg)
{
  /* cannot do much in ISR context except forward information to main loop */
  OS_API_ISetEventFlg(HOST_APL_EVENT_FLG_A, USER_INTERRUPT_BIT);
}

/* Callback function that is called when the application must be initialized */
int application_init_callback(void)
{
  GPIO_Pin_Configuration_t configuration;
  enum ble_api_result result;
  int ret;

  ret = read_config();
  if (ret) {
    dprintf("read_config() failed\n");
    return -1;
  }

  /* set up button handler */
  memset(&configuration, 0, sizeof(configuration));
  configuration.interrupt_detection_setting = RISING_EDGE;
  configuration.pull_down_resistor_setting = ON;
  configuration.pull_up_resistor_setting = OFF;
  configuration.io_configuration = INPUT;

  ret = Driver_GPIO.Initialize(GPIO4, &configuration, button_interrupt);
  if (ret) {
    warning("Driver_GPIO.Initialize() failed\n");
  }

  /* set up led handler */
  memset(&configuration, 0, sizeof(configuration));
  configuration.io_configuration = OUTPUT;
  configuration.drive_current = CURRENT_1000UA;

  ret = Driver_GPIO.Initialize(GPIO10, &configuration, NULL);
  if (ret) {
    warning("Driver_GPIO.Initialize() failed\n");
  }

  client = &doorbell_profile;

  // clang-format off
  struct ble_api_legacy_connection_parameters legacy_connection_parameters = {
    0x06, // LE_CONNECTION_INTERVAL_MINIMUM,
    0x06, // LE_CONNECTION_INTERVAL_MAXIMUM,
    0x00, // LE_CONNECTION_LATENCY,
    100,  // LE_SUPERVISION_TIMEOUT,
    LE_MINIMUM_CE_LENGTH,
    LE_MAXIMUM_CE_LENGTH,
    LE_INITIATOR_FILTER_POLICY
  };
  // clang-format on

  result = ble_api_configure_legacy_connection_parameters(&legacy_connection_parameters);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_configure_legacy_connection_parameters() failed\n");
    return -1;
  }

  /* cannot start scanning here yet */

  dprintf("\n");
  return 0;
}

/* Callback function that is peridically called when the device is awake so that application processing can be done */
int application_process_callback(void)
{
  static int once = 0;
  enum ble_api_result result;

  if (!once) {
    once = 1;

    /* on the very first call the device configuration needs to be finished. this cannot be done before.
    in the ble_application_mng_config_connection_count_response_callback() function the application processing will then be started */
    result = ble_api_mng_config_connection_count(0, 1);
    if (result) {
      dprintf("ble_api_mng_config_connection_count() failed\n");
      return -1;
    }

    return 0;
  }

  /* no further application processing is necessary */

  //  dprintf("\n");
  return 0;
}

/* Callback function of BLE API that indicates that the configuration of the role has been done. This is called once after start up and here the application processing basically starts. */
void ble_application_mng_config_connection_count_response_callback(uint8_t status)
{
  enum bt_api_result result;

  dprintf("status %02x\n", status);

  /* this is an unexpected fatal application error that should never happen */
  assert(!status);

  /* explicitely configure tx power */
  result = bt_api_ven_set_tx_power(BT_VEN_CMD_TRANSMIT_POWER_0DBM, 0, 1);
  if (result != BT_RESULT_SUCCESS) {
    dprintf("bt_api_ven_set_tx_power() failed\n");
    return;
  }
}

void bt_application_ven_set_tx_power_event_callback(uint8_t status)
{
  int ret;

  dprintf("status 0x%x\n", status);

  /* application logic here */
  ret = scan_start();
  /* this is an unexpected fatal application error that should never happen on device startup */
  assert(ret);
  (void)(ret);

  /* print out the role and the device name once as a feedback to the user */
  xprintf("CENTRAL %s\n", config.device_name);
}

/* Callback function that is called when scanning is started or stopped */
void ble_application_master_scan_status_update_callback_implementation(enum ble_api_scan_operation operation, uint8_t status)
{
  enum ble_api_result result;

  if (status) {
    dprintf("status %02x\n", status);
    fail_handler();
    return;
  }

  /* trace the current status of scanning here */

  if (operation == SCAN_START) {
    dprintf("SCAN_START\n");
    data.is_scanning = true;
    return;
  }

  /* operation == SCAN_STOP  */
  data.is_scanning = false;

  dprintf("SCAN_STOP - connecting to " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_LE(data.bd_addr));

  /* try to connect to the remote device here */
  result = ble_api_master_connect(data.bd_addr, data.addr_type, SMP_ADDRESS_TYPE_PUBLIC_ADDRESS, 5);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_master_connect() failed\n");
    fail_handler();
    return;
  }
}

enum ble_api_result ble_application_master_scan_status_update_callback(enum ble_api_scan_operation operation, uint8_t status)
{
  ble_application_master_scan_status_update_callback_implementation(operation, status);
  /* return code is ignored by BLE API... */
  return LE_API_SUCCESS;
}

void ble_application_master_advertise_report_event_callback_implementation(struct tcu_le_mng_adv_report_event *event)
{
  struct ble_api_extract_adv_data extract_adv_data_type;
  enum ble_api_result result;
  int ret;

  /* it may happen that we already requested a scan stop, but advertise reports will come in. ignore them */
  if (data.is_name_found == true && data.is_uuid_found == true) {
    return;
  }

  /* if we have been given a (device) name, this is the name we need to check
  the found device against. only if it matches we will connect to that device. */

  result = ble_api_extract_uuid_data(LE_LONG_UUID, event->adv_data, event->adv_data_len, &extract_adv_data_type);
  if (result == LE_API_SUCCESS) {
    dprintf("UUID @ " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_BE(event->address));
    ret = memcmp(client_service_doorbell.rom.uuid, extract_adv_data_type.data, LE_LONG_UUID);
    if (!ret) {
      data.is_uuid_found = true;
    }
  }

  result = ble_api_extract_device_name(event->adv_data, event->adv_data_len, &extract_adv_data_type);
  if (result == LE_API_SUCCESS) {
    dprintf("NAME '%.*s' @ " BD_ADDR_FORMAT_STRING "\n", extract_adv_data_type.data_len, extract_adv_data_type.data, BD_ADDR_ARGUMENTS_BE(event->address));
    if (extract_adv_data_type.data_len == config.device_name_len) {
      ret = memcmp(config.device_name, extract_adv_data_type.data, config.device_name_len);
      if (!ret) {
        data.is_name_found = true;
      }
    }
  }

#pragma message("Warning: This is too naive. The Bluetooth address must be stored and taken into consideration as well.")
  if (data.is_name_found == false || data.is_uuid_found == false) {
    return;
  }

  /* store peer information for later connect */
  memcpy(&data.bd_addr[0], event->address, 6);
  data.addr_type = event->address_type;

  /* we found our device - disable scanning now. processing is continued when confirmation callback comes in */
  result = ble_api_master_scan_disable();
  if (result) {
    dprintf("ble_api_master_scan_disable() failed\n");
    fail_handler();
    return;
  }

  dprintf("\n");
}

enum ble_api_result ble_application_master_advertise_report_event_callback(struct tcu_le_mng_adv_report_event *event)
{
  ble_application_master_advertise_report_event_callback_implementation(event);
  /* return code is ignored by BLE API... */
  return LE_API_SUCCESS;
}

/* Callback function that is called whenever an AT command must be processed */
int application_at_command_callback(uint8_t *buf, uint16_t len)
{
  enum ble_api_result result;
  int ret;

  /* STUNE functionality */
  if (len == 7 && buf[0] == 'S' && buf[1] == 'T' && buf[2] == 'U' && buf[3] == 'N' && buf[4] == 'E' && buf[5] == '=') {

    uint8_t tune = buf[6] - '0';

    result = ble_api_cli_characteristic_write_value_start(&characteristic_current_tune_definition, &tune, sizeof(tune), data.connection_handle);
    if (result) {
      dprintf("ble_api_cli_characteristic_write_value_start() failed\n");
      xprintf("FAILED\n");
      return 0;
    }

    dprintf("tune %02x\n", tune);
    return 0;
  }

  /* SNAME functionality */
  if (len >= 7 && buf[0] == 'S' && buf[1] == 'N' && buf[2] == 'A' && buf[3] == 'M' && buf[4] == 'E' && buf[5] == '=') {

    uint8_t *name = &buf[6];
    uint32_t name_len = len - 6;

    if (name_len > MAX_DEVICE_NAME_LEN) {
      xprintf("FAIL\n");
      return -1;
    }

    memcpy(&config.device_name[0], name, name_len);
    config.device_name_len = name_len;
    config.device_name[config.device_name_len] = 0; /* add trailing \0 */

    ret = write_config();
    if (ret) {
      xprintf("FAIL\n");
      return -1;
    }

    xprintf("OK\n");
    return 0;
  }

  return -1;
}
