/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

#ifndef __APPLICATION_H__
#define __APPLICATION_H__

/**
 * @file application.h
 * @brief Header file for the common application parts shared between the peripheral and central demo
 * @details This header files contains the common application parts that are
 * shared between the peripheral and central demo.
 *
 **/

#include "types.h"
#include "compat.h"
#include "ble_api.h"
#include "debug.h"

/* We don't want to modify types.h because it is from the SDK, so additional includes go here */
#include <stdbool.h>

/*! Definition of debugger break macro that comes from Windows. Replace it with a Cortex-M0 intrinsic */
#define __debugbreak(...) __asm("bkpt #0\n");

/* should be in host_api_add.h instead */
#define PWM_INTERRUPT_BIT (1 << 11)

/* device config in flash is stored in last flash block */
#define CONFIG_FLASH_OFFSET ((128 * 1024) - 4096)

/* Maximum length of a device name */
#define MAX_DEVICE_NAME_LEN 16

/**
 * @brief Function that reads the config block from flash.
 *
 * @param[in] buf Points to the buffer for the data to be read
 * @param[in] len Length of data buffer
 *
 * @return 0 if success, -1 if error
 */
int application_read_config(void *buf, uint32_t len);

/**
 * @brief Function that writes the config block to flash.
 *
 * @param[in] buf Points to the buffer for the data to be written
 * @param[in] len Length of data buffer
 *
 * @return 0 if success, -1 if error
 */
int application_write_config(void *buf, uint32_t len);

/**
 * @brief Function that initialises the application.
 *
 * @return 0 if success, -1 if error
 */
int application_init_callback(void);

/**
 * @brief Function that handles regular application processing. It is called whenever the
 * application has been woken up, either due to Bluetooth handling, UART handling or application
 * handling.
 *
 * @return 0 if success, -1 if error
 */
int application_process_callback(void);

/**
 * @brief Callback function that is called when an AT command needs processing.
 *
 * @param[in] buf Points to the AT command in question - AT prefix is stripped
 * @param[in] len Length of data buffer
 *
 * @return 0 if success, -1 if error
 */
int application_at_command_callback(uint8_t *buf, uint16_t len);

/*! Declaration of global server profile, used by ble_api_message_handler() */
extern struct ble_api_srv_profile *server;
/*! Declaration of global client profile, used by ble_api_message_handler() */
extern struct ble_api_cli_profile *client;

/*! Definition of a format string for printing a Bluetooth device address */
#define BD_ADDR_FORMAT_STRING "%02x:%02x:%02x:%02x:%02x:%02x"

/*! Definition of an argument list for printing a Bluetooth device address in little-endian format */
#define BD_ADDR_ARGUMENTS_BE(x) x[5], x[4], x[3], x[2], x[1], x[0]

/*! Definition of an argument list for printing a Bluetooth device address in big-endian format */
#define BD_ADDR_ARGUMENTS_LE(x) x[0], x[1], x[2], x[3], x[4], x[5]

#endif