/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

//#define DEBUG

#include "application.h"
#include "ble_api.h"
#include "system.h"
#include "osal.h"
#include "host_api.h"
#include "nvm_api_flash.h"
#include "board.h"
#include "uart1_function.h"
#include "debug.h"

/* This does not belong here */
uint8_t init_done = 0;

/* the following is TC35680 specific and probably should not be here */
#include "internal_apis.h"

struct ble_api_srv_profile *server;
struct ble_api_cli_profile *client;

struct nvm_partition system_partitions[] = {
    {nvm_reserved, 2},
    {nvm_bd_address, 6},
    {nvm_reserved, 4088},
    {
     nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition other_partitions[] = {
    {nvm_user_area, 20 * 4096}, /* User Area */
    {nvm_reserved, 4096},
    {
     nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition ble_partitions[] = {
    {nvm_reserved, 4096},            /* A-side header */
    {nvm_application_a0, 7 * 4096},  /* A0 application*/
    {nvm_application_a1, 12 * 4096}, /* A1 application*/
    {nvm_reserved, 4096},            /* B-side header */
    {nvm_application_b0, 7 * 4096},  /* B0 application*/
    {nvm_application_b1, 12 * 4096}, /* B1 application*/
    {nvm_key_store_data_1, 4096},    /* Key store data 1 */
    {nvm_key_store_data_2, 4096},    /* Key store data 2 */
    {
     nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition *const nvm_partitions[] = {
    &system_partitions[0],
    &ble_partitions[0],
    &other_partitions[0],
    NULL, /* end marker - don't remove */
};

/*! Definition of the maximum size of a ringbuffer instance */
#define RING_SIZE 512
/*! Typedef for a position in the ringbuffer. Should be the "native" type of the processor */
typedef uint32_t ring_pos_t;

/*! Structure for a ringbuffer */
struct ringbuffer {
  volatile ring_pos_t ring_head;         /*!< "Pointer" to the head of the rinbuffer */
  volatile ring_pos_t ring_tail;         /*!< "Pointer" to the tail of the ringbuffer */
  volatile uint8_t ring_data[RING_SIZE]; /*!< Data buffer for the ringbuffer */
};

/* This is a simple ringbuffer implementation. It is mainly used to store any
"keypresses" that are received from the UART for the later forwarding to an
existing Bluetooth connection. */
static void flush_rb(struct ringbuffer *rb)
{
  memset(rb, 0, sizeof(*rb));
}

static int add_to_rb(struct ringbuffer *rb, uint8_t c)
{
  ring_pos_t next_head = (rb->ring_head + 1) % RING_SIZE;
  if (next_head != rb->ring_tail) {
    /* there is room */
    rb->ring_data[rb->ring_head] = c;
    rb->ring_head = next_head;
    return 0;
  }

  /* no room left in the buffer */
  return -1;
}

static int remove_from_rb(struct ringbuffer *rb, uint8_t *c)
{
  ring_pos_t ring_head = rb->ring_head;

  if (ring_head != rb->ring_tail) {
    *c = rb->ring_data[rb->ring_tail];
    rb->ring_tail = (rb->ring_tail + 1) % RING_SIZE;
    return 0;
  }

  return -1;
}

/*! Structure that contains the application specific data */
struct application {
  struct ringbuffer in;   /*!< Ringbuffer for the intermediate storage of keypresses */
  bool linefeed_received; /*!< Indicates if a linefeed was received */
};

/*! Declaration of the global application data structure */
struct application application;

int application_read_config(void *buf, uint32_t len)
{
  uint32_t address = CONFIG_FLASH_OFFSET;
  int ret;

  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageRead(address, len, buf);
  if (ret) {
    dprintf("SYS_API_QSPIStorageRead() failed\n");
    SYS_API_QSPIStorageClose();
    return -1;
  }

  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }

  return 0;
}

int application_write_config(void *buf, uint32_t len)
{
  uint32_t address = CONFIG_FLASH_OFFSET;

  int ret;

  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageSectorErase(address);
  if (ret) {
    dprintf("SYS_API_QSPIStorageSectorErase() failed\n");
    SYS_API_QSPIStorageClose();
    return -1;
  }

  ret = SYS_API_QSPIStorageWrite(address, len, buf);
  if (ret) {
    dprintf("SYS_API_QSPIStorageWrite() failed\n");
    SYS_API_QSPIStorageClose();
    return -1;
  }

  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }

  return 0;
}

/* This function reads the pre-programmed Bluetooth device address from the
built-in flash. Surprisingly, this also initializes the Bluetooth core with
this address. So this procedure is mandatory to be executed. */
static int init_bd_address(void)
{
#define BD_ADDR_LEN 0x06
#define BD_ADDR_OFFSET 0x02
  uint8_t bd_address[BD_ADDR_LEN];
  int ret;

  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageReadSpecifiedAddr(BD_ADDR_LEN, BD_ADDR_OFFSET);
  if (ret) {
    dprintf("SYS_API_QSPIStorageReadSpecifiedAddr() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageRead(BD_ADDR_OFFSET, BD_ADDR_LEN, &bd_address[0]);
  if (ret) {
    dprintf("SYS_API_QSPIStorageRead() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

__weak void button_handler_callback(void)
{
}

__weak void pwm_handler(void)
{
}

static int common_process(void)
{
#define MAX_COMMAND_LENGTH 32
  uint8_t buf[MAX_COMMAND_LENGTH + 1];
  int i;
  int ret;

  //  dprintf("data_in_rb %d\n", data_in_rb(&application.in));

  if (application.linefeed_received == false) {
    return 0;
  }

  application.linefeed_received = false;

  for (i = 0; i < sizeof(buf) - 1; i++) {
    ret = remove_from_rb(&application.in, &buf[i]);
    if (ret) {
      xprintf("error while reading command from ringbuffer\n");
      return -1;
    }
    if (buf[i] == '\n') {
      break;
    }
  }

  flush_rb(&application.in);

  if (i == sizeof(buf) - 1) {
    xprintf("no command found in ringbuffer\n");
    return -1;
  }
  buf[i] = 0;

  if (buf[0] != 'A' && buf[1] != 'T') {
    //    xprintf("AT command signature not found\n");
    return -1;
  }

  ret = application_at_command_callback(&buf[2], i - 2);
  if (!ret) {
    return 0;
  }

  xprintf("UNKNOWN\n");
  return 0;
}

/* This is the main function. It does the basic initialization and then goes
into the main message processing loop. The "system" and "message" handling
functions come straight from the Bluetooth SDK. */
C2_ENTRY int C2_Task(void)
{
  enum bt_tsb_bluetooth_chip_version version;
  enum ble_api_result result;
  unsigned int flag = 0;
  uint16_t pmu_cause;
  uint16_t crw_rst_cause;
  int ret;

  DATA_INIT();

  uart1_init();

  ret = SYS_API_GetWakeUpCauseValue(NULL, &pmu_cause, &crw_rst_cause);
  if (ret) {
    dprintf("SYS_API_GetWakeUpCauseValue() failed\n");
    goto fatal_error;
  }

  if (pmu_cause & 0x01) {
    dprintf("wake up from deep sleep\n");
  } else if (pmu_cause & (0x01 << 3)) {
    dprintf("wake up from cold boot\n");
  } else {
    dprintf("unknown wake up reason\n");
  }

  ret = init_bd_address();
  if (ret) {
    dprintf("init_bd_address() failed\n");
    goto fatal_error;
  }

  ret = system_read_version(&version);
  if (ret) {
    dprintf("system_read_version() failed\n");
    goto fatal_error;
  }

  /* TC35678 requires fw patches. This function is used for this purpose */
  C2_Patch();  
  
  /* do not set up low power mode - UART 1 does not wake up system */

  result = ble_api_configure_bt_chip_mode(version, LEGACY);
  if (result) {
    vPrintf("Mode Init failed\n");
    goto fatal_error;
  }

  result = ble_api_init(version);
  if (result) {
    dprintf("ble_api_init() failed\n");
    goto fatal_error;
  }

  init_done |= BLE_INIT;

  ret = nvm_init(NVM_SIZE);
  if (ret) {
    dprintf("nvm_init() failed\n");
    goto fatal_error;
  }

  ret = key_store_ble_init();
  if (ret) {
    dprintf("key_store_ble_init() failed\n");
    goto fatal_error;
  }

  /* The peripheral and central demo need different init functions */
  ret = application_init_callback();
  if (ret) {
    dprintf("application_init_callback() failed\n");
    goto fatal_error;
  }

  while (1) {

    common_process();
    application_process_callback();

    event_wait(&flag);

    /* we do not handle any low power events here */
    /* we do not handle any rtc events here either */

    if ((flag & C2_MSG_INT_BIT)) {
      /* ble_api_message_handler is called if received event is BLE event */
      ret = ble_api_message_handler(server, client);
      if (ret) {
        dprintf("ble_api_message_handler() failed\n");
        /* fall through */
      }
    }
    if ((flag & USER_INTERRUPT_BIT)) {
      button_handler_callback();
    }
    if ((flag & PWM_INTERRUPT_BIT)) {
      pwm_handler();
    }
  }

fatal_error:
  warning("fatal error - application did not start\n");
  while (1) {
    /* fixme */
  }
}

int uart1_receive_bytes_callback_isr(uint8_t *ptr, int len)
{
  int i;
  int ret;

  for (i = 0; i < len; i++) {
    ret = add_to_rb(&application.in, ptr[i]);
    if (ret) {
      /* if ringbuffer is full and no linefeed in the buffer, clean it */
      if (application.linefeed_received == false) {
        flush_rb(&application.in);
      }
      return -1;
    }
    if (ptr[i] == '\n') {
      application.linefeed_received = true;
    }
  }

  return 0;
}

#include "mini-printf.h"

static uint8_t buf[128];
int debug_send(const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = mini_vsnprintf((char *)buf, sizeof(buf), format, args);
  va_end(args);

  uart1_send(buf, ret);
  return 0;
}

void data_receive_debug_callback(uint8_t *buf, uint32_t length)
{
#if 0
  uint32_t i;

  for (i = 0; i < length; i++) {
    /* This will print raw data in ASCII format.
     for binary data can use %c format specifier */
    debug_send("0x%x ", buf[i]);
  }

  debug_send("\n");
#endif
}

void tcu_message_debug_output(const char *prefix, const uint8_t *buf, uint32_t length)
{
#if 0
  static char debug_buf[256];
  
  int i;
  char *ptr = &debug_buf[0];
  int len = sizeof(debug_buf);
  int ret;
  
  ret = mini_snprintf(ptr, len, "%03d %s ", length, prefix);
  ptr += ret; len -= ret;
  
  for (i = 0; i < length; i++) {
    ret = mini_snprintf(ptr, len, "%02x ", buf[i]);
    ptr += ret;  len -= ret;
    if (!len) {
      break;
    }
  }

  uart1_printf("%s\n", debug_buf);
#endif
}

void ble_application_default_event_callback(struct bt_cmn_msg *cmn_msg, uint8_t *data)
{
  warning("\n");
}

void ble_application_uart_message_callback(struct bt_cmn_msg *cmn_msg, uint8_t *data)
{
  application_process_callback();
}

extern int message_timer_callback(void *timer);

void ble_application_timer_message_callback(struct bt_cmn_msg *cmn_msg, uint8_t *data)
{
  message_timer_callback((void *)cmn_msg->user_data);
}

void ble_application_event_handling_failed_callback(uint8_t service_id, uint8_t opcode, uint8_t error)
{
  xprintf("service_id %02x, opcode %02x, error %02x\n", service_id, opcode, error);
  __debugbreak();
}

void ble_application_mng_phy_update_event_callback(uint8_t status, uint16_t connection_handle, uint8_t tx_phy, uint8_t rx_phy)
{
  dprintf("status 0x%x, connection_handle 0x%x, tx_phy 0x%x, rx_phy 0x%x\n", status, connection_handle, tx_phy, rx_phy);
  if (tx_phy == 0x01) {
    xprintf("TX LE 1M\n");
  } else if (tx_phy == 0x02) {
    xprintf("TX LE 2M\n");
  } else if (tx_phy == 0x03) {
    xprintf("TX LE Coded\n");
  }
  if (rx_phy == 0x01) {
    xprintf("RX LE 1M\n");
  } else if (rx_phy == 0x02) {
    xprintf("RX LE 2M\n");
  } else if (rx_phy == 0x03) {
    xprintf("RX LE Coded\n");
  }
}

int ble_application_mng_update_connection_request_callback(uint16_t connection_handle, struct ble_connection_update_req_event_params *update_parameters)
{
  struct ble_connection_update_accept_params update_request_params;

  update_request_params.conn_interval_min = update_parameters->conn_interval_min;
  update_request_params.conn_interval_max = update_parameters->conn_interval_max;
  update_request_params.conn_latency = update_parameters->conn_latency;
  update_request_params.sup_time_out = update_parameters->sup_time_out;

  /* automatically accept any incoming update connection request */
  ble_api_remote_connection_parameter_accept_req(connection_handle, LE_UPDATE_ACCEPTED, 0, &update_request_params);
  dprintf("connection_handle %04x\n", connection_handle);
  return 0;
}

int ble_application_mng_remote_connection_parameter_request_callback(uint16_t connection_handle, struct ble_connection_update_req_event_params *update_parameters)
{
  struct ble_connection_update_accept_params update_request_params;

  update_request_params.conn_interval_min = update_parameters->conn_interval_min;
  update_request_params.conn_interval_max = update_parameters->conn_interval_max;
  update_request_params.conn_latency = update_parameters->conn_latency;
  update_request_params.sup_time_out = update_parameters->sup_time_out;

  /* automatically accept any incoming remote update connection request */
  ble_api_remote_connection_parameter_accept_req(connection_handle, LE_UPDATE_ACCEPTED, 0, &update_request_params);
  dprintf("connection_handle %04x\n", connection_handle);
  return 0;
}

#include "ble_api_smp.h"

int ble_application_smp_pairing_request_callback(enum ble_smp_device_role role, uint16_t connection_handle, enum tcu_le_smp_authentication_requirements authentication_requirement)
{
  dprintf("connection_handle %04x, role %02x, authentication_requirement %02x\n", connection_handle, role, authentication_requirement);

  if (role == SMP_SLAVE) {
    ble_api_smp_slv_accept_pairing_event(connection_handle, authentication_requirement);
  } else {
    ble_api_smp_master_accept_pairing_request(connection_handle, authentication_requirement);
  }

  return 0;
}

void ble_application_smp_status_callback(enum tcu_le_smp_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %02x\n", connection_handle, status);

  if (status == STATUS_SUCCESS) {
    xprintf("ENCRYPTION_OK\n");
    return;
  }

  xprintf("ENCRYPTION_FAILED\n");
}

void ble_application_smp_key_distribution_callback(uint8_t *initiator_key, uint8_t *responder_key)
{
  *initiator_key = SMP_LTK_ENABLED | SMP_IRK_ENABLED;
  *responder_key = SMP_LTK_ENABLED | SMP_IRK_ENABLED;
  dprintf("\n");
}

void ble_application_smp_capability_callback(enum tcu_le_smp_io_capabilities *io_capabilities)
{
  *io_capabilities = SMP_IO_CAP_NO_IP_NO_OP;
  dprintf("\n");
}

void ble_application_smp_oob_data_available_callback(uint16_t connection_handle, enum tcu_le_smp_slv_mas_oob_authentication_data_available *own_flag, enum tcu_le_smp_slv_mas_oob_authentication_data_available *remote_flag)
{
  *own_flag = SMP_OOB_DATA_NOT_PRESENT;
  *remote_flag = SMP_OOB_DATA_NOT_PRESENT;
  dprintf("\n");
}

#include "bt_api_ven_flash.h"

int board_nvm_open_flash(uint32_t write_protection)
{
  return bt_api_ven_nvm_storage_open(write_protection);
}

int board_nvm_close_flash(void)
{
  return bt_api_ven_nvm_storage_close();
}

int board_nvm_read_flash(uint32_t address, uint8_t *data, uint32_t length)
{
  return bt_api_ven_nvm_storage_read(address, data, (uint16_t)length);
}

int board_nvm_sector_erase_flash(uint32_t address)
{
  return bt_api_ven_nvm_storage_sector_erase(address);
}

int board_nvm_write_flash(uint32_t address, uint8_t *data, uint32_t length)
{
  return bt_api_ven_nvm_storage_write(address, data, (uint16_t)length);
}
