/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/
//#define DEBUG

#include "application.h"
#include "ble_api_slave.h"
#include "bt_api_ven_flash.h"
#include "crc8.h"
#include "debug_api.h"

#include "characteristic_gap_appearance.h"
#include "characteristic_gap_device_name.h"
#include "characteristic_hardware_revision_string.h"
#include "characteristic_manufacturer_name_string.h"
#include "characteristic_model_number_string.h"
#include "characteristic_scan_interval_window.h"
#include "characteristic_scan_refresh.h"
#include "characteristic_system_id.h"

#include "music.h"

/*! Structure that contains peripheral-specific application data */
struct data {
  bool signal_music_started;  /*!< signals if the music has been started */
  bool signal_music_stopped;  /*!< signals if the music has been stopped */
  uint16_t connection_handle; /*!< connection handle of the currently active connection */
};
static struct data data;

/*! Structure that contains configuration data for the peripheral */
struct config {
  uint8_t device_name[MAX_DEVICE_NAME_LEN + 1]; /*!< device name used for advertising */
  uint8_t device_name_len;                      /*!< length of device name */
  uint8_t current_tune_characteristic_value;    /*!< characteristic value of the current tune characteristic */
  uint8_t crc8;                                 /*!< crc value - must be last */
};
static struct config config;

/* Forward declaration of helper function */
static int write_config(void);

/* Device Information Service configuration data */
static const uint8_t manufacturer_name[] = "Panasonic";
static const uint8_t model_number[] = "PAN-USB-ETU";
static const uint8_t manufacturer_identifier[] = {0x01, 0x02, 0x03, 0x04, 0x05};
static const uint8_t organizationally_unique_identifier[] = {0x01, 0x02, 0x03};
static const uint8_t hw_revision_string[] = "TOSHIBA";

/* Doorbell characteristics */

/* "Ring" characteristic */

static uint8_t is_ringing; /*!< characteristic value of the ring characteristic */

/* Callback function that is called when ring characteristic value is initialized */
static enum tcu_le_gatt_status init_characteristic_ring(uint8_t **value, uint16_t *length)
{
  *length = sizeof(is_ringing);
  *value = &is_ringing;
  dprintf("length %d\n", *length);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when ring characteristic value is read */
static enum tcu_le_gatt_status read_characteristic_ring(uint8_t **value, uint16_t *length, uint16_t connection_handle)
{
  *length = sizeof(is_ringing);
  *value = &is_ringing;
  dprintf("connection_handle %04x, length %d\n", connection_handle, *length);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when ring characteristic value is about to send via indication */
static enum tcu_le_gatt_status indicate_characteristic_ring(uint8_t **value, uint16_t *length)
{
  *length = sizeof(is_ringing);
  *value = &is_ringing;
  dprintf("length %d\n", *length);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when indication has been sent (or not in case of failure) */
static int indicate_done_characteristic_ring(enum tcu_le_gatt_status status, uint16_t connection_handle)
{
  dprintf("connection_handle %04x, status %d\n", connection_handle, status);

  /* In this application there is no need to handle an error when sending indiciations */
  if (status != LE_GATT_STATUS_SUCCESS) {
    return 0;
  }

  return 0;
}

/* Callback function that is called when ring characteristic value has been written */
static enum tcu_le_gatt_status write_characteristic_ring(uint8_t *value, uint16_t length, uint16_t connection_handle)
{
  int ret;

  if (length != 1) {
    dprintf("length %d out of bounds\n", length);
    return LE_GATT_STATUS_SUCCESS;
  }

  /* start the music regardless of the value that is written - value is not stored here either */
  ret = music_start();
  if (ret) {
    dprintf("music_start() failed\n");
    /* fixme: return an error code here instead? */
    return LE_GATT_STATUS_SUCCESS;
  }

  dprintf("connection_handle %04x, length %d\n", connection_handle, length);

  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when ring characteristic value is about to be written */
static enum tcu_le_gatt_status prep_write_characteristic_ring(uint16_t offset, uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when ring client characteristic configuration descriptor is about to be written */
static enum tcu_le_gatt_status client_characteristic_configuration_descriptor_ring_prep_write(uint16_t offset, uint16_t connection_handle)
{
  dprintf("connection_handle %04x\n", connection_handle);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when ring client characteristic configuration descriptor has been written */
static enum tcu_le_gatt_status client_characteristic_configuration_descriptor_ring_write_done(uint16_t connection_handle)
{
  /* no special handling is necessary when indications are enabled or disabled */
  dprintf("connection_handle %04x\n", connection_handle);
  return LE_GATT_STATUS_SUCCESS;
}

/* Forward declaration that is necessary to be able to declare a client characteristic configuration descriptor for the ring characteristic */
extern struct ble_api_srv_characteristic characteristic_ring_definition;

/* Declaration of a client characteristic configuration descriptor for the ring characteristic */
BLE_API_SERVER_CLIENT_CHARACTERISTIC_CONFIGURATION_DESCRIPTOR(ring, 0x0074)

/* Array of all characterstic descriptors for the ring characteristic - there is only one */
static struct ble_api_srv_characteristic_descriptor *characteristic_ring_characteristic_descriptors[] = {
    &client_characteristic_configuration_descriptor_ring,
};

/* Total number of characterstic descriptors for the ring characteristic */
#define NUM_CHARACTERISTIC_RING_CHARACTERISTIC_DESCRIPTORS (sizeof(characteristic_ring_characteristic_descriptors) / sizeof(characteristic_ring_characteristic_descriptors[0]))

/* Declaration of storage for runtime data of the ring characteristic */
static struct ble_api_srv_characteristic_ram characteristic_ring_definition_ram;

/* Declaration of the ring characteristic */
static struct ble_api_srv_characteristic characteristic_ring_definition = {
    {
        LE_LONG_UUID,
        /* UUID 49700b78-1e46-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
            0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x46, 0x1e, 0x78, 0x0b, 0x70, 0x49,
            // clang-format on
        },
        LE_GATT_CHAR_FIXED_LENGTH_TYPE,
        (LE_READ | LE_WRITE | LE_INDICATE),
        {
            init_characteristic_ring,
            NULL,
            read_characteristic_ring,
            NULL,
            indicate_characteristic_ring,
            prep_write_characteristic_ring,
            write_characteristic_ring,
            indicate_done_characteristic_ring,
            NULL,
            NULL,
            0x0073,
        },
        0,
        NUM_CHARACTERISTIC_RING_CHARACTERISTIC_DESCRIPTORS,
        &characteristic_ring_characteristic_descriptors[0],
        NULL,
        0x0072,
    },
    &characteristic_ring_definition_ram,
};

/* "Current tune" characteristic */

/* The "current tune" characteristic value is stored persistently in the configuration data */

/* Callback function that is called when current tune characteristic value is initialized */
enum tcu_le_gatt_status init_characteristic_current_tune(uint8_t **value, uint16_t *length)
{
  *length = sizeof(config.current_tune_characteristic_value);
  *value = &config.current_tune_characteristic_value;
  dprintf("length %d\n", *length);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when current tune characteristic value is read */
enum tcu_le_gatt_status read_characteristic_current_tune(uint8_t **value, uint16_t *length, uint16_t connection_handle)
{
  *length = sizeof(config.current_tune_characteristic_value);
  *value = &config.current_tune_characteristic_value;
  dprintf("connection_handle %04x, length %d\n", connection_handle, *length);
  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when current tune characteristic value is about to be written */
static enum tcu_le_gatt_status prep_write_characteristic_current_tune(uint16_t offset, uint16_t connection_handle)
{
  return LE_GATT_STATUS_SUCCESS;
}

/*! Helper function that translates a value into a doorbell tune identifier */
static int set_music_tune(int value)
{
  switch (value) {
    case 0:
      music_tune = music_tune_fuer_elise;
      break;
    case 1:
      music_tune = music_tune_ode_an_die_freude;
      break;
    case 2:
      music_tune = music_tune_kleine_nachtmusik;
      break;
    default:
      dprintf("value %02x out of bounds\n", value);
      return -1;
  }

  dprintf("value %02x\n", value);
  return 0;
}

/* Callback function that is called when current tune characteristic value is written */
static enum tcu_le_gatt_status write_characteristic_current_tune(uint8_t *value, uint16_t length, uint16_t connection_handle)
{
  int ret;

  if (length != 1) {
    dprintf("length %d out of bounds\n", length);
    return LE_GATT_STATUS_INVALID_PARAMETER_LENGTH;
  }

  /* we change the current tune characteristic value here - the value will be used when then next ring tone is played */
  ret = set_music_tune(*value);
  if (ret) {
    dprintf("set_music_tune() failed\n");
    return LE_GATT_STATUS_INVALID_REQUEST;
  }

  /* store the value in the configuration data as well */
  config.current_tune_characteristic_value = *value;

  /* and update the configuration data in flash */
  ret = write_config();
  if (ret) {
    dprintf("write_config() failed\n");
    /* fall through - there is nothing we can do here */
  }

  dprintf("value %02x\n", *value);

  return LE_GATT_STATUS_SUCCESS;
}

/* Callback function that is called when current tune characteristic value was written */
static enum tcu_le_gatt_status write_done_characteristic_current_tune(uint16_t connection_handle)
{
  return LE_GATT_STATUS_SUCCESS;
}

/* Valid range descriptor value of the current tune characteristic */
static uint8_t valid_range[2] = {0, 2}; /* the allowed range is from 0 to 2 */

static enum tcu_le_gatt_status current_tune_valid_range_descriptor_read(uint8_t **value, uint16_t *length, uint16_t connection_handle)
{
  /* During intialization as no connection handle is availabe, dummy connection hadnle 0x0000 is passed. Thus it is made 0x0001. During run-time actual connection handle must be passed */
  if (connection_handle == 0x0000) {
    connection_handle = 0x0001;
  }

  *value = &valid_range[0];
  *length = sizeof(valid_range);

  dprintf("\n");

  return LE_GATT_STATUS_SUCCESS;
}

/* Declaration of a valid range descriptor for the current tune characteristic */
static struct ble_api_srv_characteristic_descriptor_ram current_tune_valid_range_descriptor_ram;

static struct ble_api_srv_characteristic_descriptor current_tune_valid_range_descriptor = {
    {
        LE_SHORT_UUID,
        SET_LE_SHORT_UUID(0x2906),
        LE_GATT_CHAR_DESC_FIXED_LENGTH_TYPE,
        (uint16_t)(LE_GATT_READ),
        NULL, /* n/a */
        NULL, /* prep write */
        NULL, /* write */
        NULL, /* write done */
        current_tune_valid_range_descriptor_read,
        NULL, /* read var desc */
        0x77,
    },
    &current_tune_valid_range_descriptor_ram,
};

/* Array of all characterstic descriptors for the ring characteristic - there is only one */
static struct ble_api_srv_characteristic_descriptor *characteristic_current_tune_characteristic_descriptors[] = {
    &current_tune_valid_range_descriptor,
};

/* Total number of characterstic descriptors for the ring characteristic */
#define NUM_CHARACTERISTIC_CURRENT_TUNE_CHARACTERISTIC_DESCRIPTORS (sizeof(characteristic_current_tune_characteristic_descriptors) / sizeof(characteristic_current_tune_characteristic_descriptors[0]))

/* Declaration of storage for runtime data of the current tune characteristic */
static struct ble_api_srv_characteristic_ram characteristic_current_tune_definition_ram;

/* Declaration of the current tune characteristic */
static struct ble_api_srv_characteristic characteristic_current_tune_definition = {
    {
        LE_LONG_UUID,
        /* UUID 82ab3028-1e48-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
          0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x48, 0x1e, 0x28, 0x30, 0xab, 0x82,
            // clang-format on
        },
        LE_GATT_CHAR_FIXED_LENGTH_TYPE,
        (LE_WRITE | LE_READ),
        {
            init_characteristic_current_tune,
            NULL, /* init_var_char */
            read_characteristic_current_tune,
            NULL, /* notify */
            NULL, /* indicate */
            prep_write_characteristic_current_tune,
            write_characteristic_current_tune,
            NULL, /* indicate_done */
            NULL, /* notify_done */
            write_done_characteristic_current_tune,
            0x0076,
        },
        0,
        NUM_CHARACTERISTIC_CURRENT_TUNE_CHARACTERISTIC_DESCRIPTORS,
        characteristic_current_tune_characteristic_descriptors,
        NULL, /* timer */
        0x0075,
    },
    &characteristic_current_tune_definition_ram,
};

/* Doorbell service */

/* Declaration of all the characteristics of the doorbell service */
static struct ble_api_srv_characteristic *service_doorbell_characteristics[] = {
    &characteristic_ring_definition,
    &characteristic_current_tune_definition,
};

/* Total number of services of the doorbell service */
#define NUM_DOORBELL_SERVICE_CHARACTERISTICS sizeof(service_doorbell_characteristics) / sizeof(service_doorbell_characteristics[0])

/* Declaration of storage for runtime data of the doorbell service  */
static struct ble_api_srv_service_ram service_doorbell_ram;

/* Declaration of the doorbell service */
struct ble_api_srv_service service_doorbell = {
    {
        LE_LONG_UUID,
        /* UUID dcace68e-1e48-11e9-ab14-d663bd873d93 */
        {
            // clang-format off
          0x93, 0x3d, 0x87, 0xbd, 0x63, 0xd6, 0x14, 0xab, 0xe9, 0x11, 0x48, 0x1e, 0x8e, 0xe6, 0xac, 0xdc,
            // clang-format on
        },
        NULL,
        NUM_DOORBELL_SERVICE_CHARACTERISTICS,
        &service_doorbell_characteristics[0],
        0x0071,
    },
    &service_doorbell_ram,
};

void ble_application_slave_advertising_status_update_callback(enum ble_api_adv_operation operation, uint8_t status)
{
  dprintf("operation %d, status %02x\n", operation, status);

  /* this is an unexpected fatal application error that should never happen */
  assert(!status);

  /* print out the role and the device name once as a feedback to the user */
  xprintf("PERIPHERAL %s\n", g_characteristic_gap_device_name.name);
}

void bt_application_ven_set_tx_power_event_callback(uint8_t status)
{
  enum ble_api_result result;

  dprintf("status %02x\n", status);

  /* this is an unexpected fatal application error that should never happen */
  assert(!status);

  /* start advertising */
  result = ble_api_peripheral_role_start(ADDRESS_CONFIGURATION_PUBLIC, LEGACY_AUTO_ADV);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_peripheral_role_start() failed\n");
    return;
  }
}

/* Callback function of BLE API that indicates that the configuration of the role has been done. This is called once after start up and here the application processing basically starts. */
void ble_application_mng_config_connection_count_response_callback(uint8_t status)
{
  enum bt_api_result result;

  dprintf("status %02x\n", status);

  /* this is an unexpected fatal application error that should never happen */
  assert(!status);

  /* explicitely configure tx power */
  result = bt_api_ven_set_tx_power(BT_VEN_CMD_TRANSMIT_POWER_0DBM, 0, 1);
  if (result != BT_RESULT_SUCCESS) {
    dprintf("bt_api_ven_set_tx_power() failed\n");
    return;
  }
}

/* Callback function that is peridically called when the device is awake so that application processing can be done */
int application_process_callback(void)
{
  static int once = 0;
  enum ble_api_result result;
  int ret;

  if (!once) {
    once = 1;

    /* on the very first call the device configuration needs to be finished. this cannot be done before.
    in the ble_application_mng_config_connection_count_response_callback() function the application processing will then be started */
    result = ble_api_mng_config_connection_count(1, 0);
    if (result) {
      dprintf("ble_api_mng_config_connection_count() failed\n");
      return -1;
    }

    dprintf("once\n");
    return 0;
  }

  /* handle the case that ringtone has started playing */
  if (data.signal_music_started == true) {

    is_ringing = 1;
    data.signal_music_started = false;

    /* if indications for the ring characteristic are enabled ... */
    ret = ble_api_srv_characteristic_is_indication_enabled(&characteristic_ring_definition, data.connection_handle);
    if (ret) {
      dprintf("indications not enabled\n");
      return 0;
    }

    /* ... then send an indication with the current characteristic value */
    ret = ble_api_srv_characteristic_changed(&characteristic_ring_definition, data.connection_handle);
    if (ret != LE_API_SUCCESS) {
      dprintf("ble_api_srv_characteristic_changed() failed\n");
      return -1;
    }

    dprintf("handling signal_music_started\n");
    return 0;
  }

  /* handle the case that ringtone has stopped playing */
  if (data.signal_music_stopped == true) {

    is_ringing = 0;
    data.signal_music_stopped = false;

    /* if indications for the ring characteristic are enabled ... */
    ret = ble_api_srv_characteristic_is_indication_enabled(&characteristic_ring_definition, data.connection_handle);
    if (ret) {
      dprintf("indications not enabled\n");
      return 0;
    }

    /* ... then send an indication with the current characteristic value */
    ret = ble_api_srv_characteristic_changed(&characteristic_ring_definition, data.connection_handle);
    if (ret != LE_API_SUCCESS) {
      dprintf("ble_api_srv_characteristic_changed() failed\n");
      return -1;
    }

    dprintf("handling signal_music_stopped\n");
    return 0;
  }

  dprintf("\n");
  return 0;
}

/* Callback function that is called when the must starts playing */
void music_started_callback(void)
{
  /* this is an asynchronous callback in the context of the music driver. we should not execute any functionality directly, but forward the request to the application process handling  */
  is_ringing = 1;
  data.signal_music_started = true;
  dprintf("\n");
}

/* Callback function that is called when the must stops playing */
void music_stopped_callback(void)
{
  /* this is an asynchronous callback in the context of the music driver. we should not execute any functionality directly, but forward the request to the application process handling  */
  is_ringing = 0;
  data.signal_music_stopped = true;
  dprintf("\n");
}

/* Doorbell profile */

extern struct ble_api_srv_service service_generic_access;
extern struct ble_api_srv_service service_device_information;

/* Declaration of all the services of the doorbell profile */
static struct ble_api_srv_service *services[] = {
    &service_generic_access,
    &service_device_information,
    &service_doorbell,
};

/* Callback function that is called when the profile is initialized */
static enum ble_api_result init(struct ble_api_srv_profile *profile)
{
  g_characteristic_gap_appearance.category = APPEARANCE_CATEGORY_UNKNOWN;
  return LE_API_SUCCESS;
}

/* Callback function that is called when the profile initialization is done */
static enum ble_api_result profile_init_done(struct ble_api_srv_profile *profile)
{
  advertising_data.server = profile;
  advertising_data.use_appearance = 0;
  advertising_data.advertising_interval_minimum = LE_ADVERTISING_INTERVAL_MINIMUM;
  advertising_data.advertising_interval_maximum = LE_ADVERTISING_INTERVAL_MAXIMUM;

  return LE_API_SUCCESS;
}

/* Callback function that is called when a connection has been established */
static void connected_callback(uint8_t *bd_addr, uint16_t connection_handle, uint16_t mtu_size)
{
  int ret;

  /* the LED is turned on to indicate the connection status */
  ret = Driver_GPIO.Write(GPIO10, HIGH);
  if (ret) {
    warning("Driver_GPIO.Write() failed\n");
  }

  /* store connection handle - for example this is used for sending indications from the application processing */
  data.connection_handle = connection_handle;

  xprintf("CONNECT " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_BE(bd_addr));

  dprintf("\n");
}

/* Callback function that is called when a connection has been terminated */
static void disconnected_callback(uint8_t *bd_addr, uint16_t connection_handle)
{
  enum ble_api_result result;
  int ret;

  xprintf("DISCONNECT " BD_ADDR_FORMAT_STRING "\n", BD_ADDR_ARGUMENTS_BE(bd_addr));

  data.connection_handle = 0;

  /* the LED is turned off to indicate the connection status */
  ret = Driver_GPIO.Write(GPIO10, LOW);
  if (ret) {
    warning("Driver_GPIO.Write() failed\n");
  }

  /* advertising is turned on for the next connection */
  result = ble_api_peripheral_role_start(ADDRESS_CONFIGURATION_PUBLIC, LEGACY_AUTO_ADV);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_peripheral_role_start() failed\n");
    return;
  }

  dprintf("\n");
}

/* Callback fundction that is called when the MTU size changes */
static void mtu_changed_callback(uint8_t *bd_addr, uint16_t connection_handle, uint16_t mtu_size)
{
  /* nothing to be done here */
}

/* Declaration of storage for runtime data of the doorbell profile  */
static struct ble_api_srv_profile_ram profile_doorbell_ram[MAXIMUM_DEVICES];

/* Declaration of the doorbell profile */
struct ble_api_srv_profile profile_doorbell = {
    {
        services,
        sizeof(services) / sizeof(services[0]),
        SMP_AUTH_NO_MITM_NO_BONDING,
        init,
        profile_init_done,
        connected_callback,
        disconnected_callback,
        mtu_changed_callback,
    },
    &profile_doorbell_ram[0],
};

/* Helper function that updates the configuration data in the flash */
static int write_config(void)
{
  int ret;

  /* Calculate a crc for data integrity */
  config.crc8 = crc((void *)&config, sizeof(config) - 1);

  ret = application_write_config(&config, sizeof(config));
  if (ret) {
    dprintf("application_write_config() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

/* Helper function that reads the configuration data from flash */
static int read_config(void)
{
  /* The default device name to use when no configuration is present */
  static const uint8_t device_name[] = "PAN1760A\0";
  uint8_t calculated_crc8;
  int ret;

  ret = application_read_config(&config, sizeof(config));
  if (ret) {
    dprintf("application_read_config() failed\n");
    return -1;
  }

  /* check if configuration data is valid by comparing a calculated crc with the stored one */
  calculated_crc8 = crc((void *)&config, sizeof(config) - 1);
  if (calculated_crc8 == config.crc8) {
    /* configuration is valid */
  } else {
    /* use default configuration */
    memcpy(&config.device_name[0], &device_name[0], sizeof(device_name));
    config.device_name_len = sizeof(device_name) - 1; /* do not count trailing \0 */
    config.current_tune_characteristic_value = 0;
  }

  /* configure the music tune from the configuration data */
  ret = set_music_tune(config.current_tune_characteristic_value);
  if (ret) {
    dprintf("music_set_tune() failed\n");
    /* there is nothing we can do here - hope for the best */
  }

  dprintf("\n");
  return 0;
}

/* Callback function that is called when the application is initialized */
int application_init_callback(void)
{
  enum ble_api_result result;
  GPIO_Pin_Configuration_t configuration;
  int ret;

  ret = read_config();
  if (ret) {
    dprintf("read_config() failed\n");
    return -1;
  }

  /* Configure GPIO for LED handling */
  memset(&configuration, 0, sizeof(configuration));
  configuration.io_configuration = OUTPUT;
  configuration.drive_current = CURRENT_1000UA;

  ret = Driver_GPIO.Initialize(GPIO10, &configuration, NULL);
  if (ret) {
    dprintf("Driver_GPIO.Initialize() failed\n");
    return -1;
  }

  ret = music_init();
  if (ret) {
    dprintf("music_init() failed\n");
    return -1;
  }

  /* gap service */
  g_characteristic_gap_device_name.length = config.device_name_len;
  g_characteristic_gap_device_name.name = config.device_name;

  /* device information service */
  g_characteristic_model_number_string.model_number = model_number;
  g_characteristic_model_number_string.length = sizeof(model_number) - 1;
  memcpy(&g_characteristic_system_id.manufacturer_identifier[0], &manufacturer_identifier[0], sizeof(g_characteristic_system_id.manufacturer_identifier));
  memcpy(&g_characteristic_system_id.organizationally_unique_identifier[0], &organizationally_unique_identifier[0], sizeof(g_characteristic_system_id.organizationally_unique_identifier));
  g_characteristic_manufacturer_name_string.manufacturer_name = (uint8_t *)manufacturer_name;
  g_characteristic_manufacturer_name_string.length = sizeof(manufacturer_name) - 1;
  g_characteristic_hardware_revision_string.hw_string = hw_revision_string;
  g_characteristic_hardware_revision_string.length = sizeof(hw_revision_string) - 1;

  advertising_data.advertising_name = config.device_name;
  advertising_data.adv_name_length = config.device_name_len;

  server = &profile_doorbell;

  result = ble_api_srv_profile_init(&profile_doorbell);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_srv_profile_init() failed\n");
    return -1;
  }

  return 0;
}

/* Callback function that is called whenever the user decides to change the device name */
void application_device_name_changed(uint8_t *name, uint16_t len)
{
  /* the device name is only used for advertising - this will take effect when the next advertising is started */
  advertising_data.advertising_name = name;
  advertising_data.adv_name_length = len;
  /* todo: stop advertising and re-start it again here */
}

/* Callback function that is called whenever an AT command must be processed */
int application_at_command_callback(uint8_t *buf, uint16_t len)
{
  int ret;

  /* SNAME functionality */
  if (len >= 7 && buf[0] == 'S' && buf[1] == 'N' && buf[2] == 'A' && buf[3] == 'M' && buf[4] == 'E' && buf[5] == '=') {

    uint8_t *name = &buf[6];
    uint32_t name_len = len - 6;

    if (name_len > MAX_DEVICE_NAME_LEN) {
      xprintf("FAIL\n");
      return -1;
    }

    memcpy(&config.device_name[0], name, name_len);
    config.device_name_len = name_len;
    config.device_name[config.device_name_len] = 0; /* add trailing \0 */

    ret = write_config();
    if (ret) {
      xprintf("FAIL\n");
      return -1;
    }

    xprintf("OK\n");
    return 0;
  }

  return -1;
}

/* Stubbing - used by central role, but not peripheral role */
#include "ble_api_master.h"

enum ble_api_result ble_application_master_advertise_report_event_callback(struct tcu_le_mng_adv_report_event *event)
{
  __debugbreak();
  return LE_API_SUCCESS;
}

enum ble_api_result ble_application_master_scan_status_update_callback(enum ble_api_scan_operation operation, uint8_t status)
{
  __debugbreak();
  return LE_API_SUCCESS;
}
